// Grid.java
// CS201J Fall 2002

import java.util.Vector;
import java.util.Enumeration;
import java.util.Random;

public class Grid {
    // OVERVIEW: A Grid is a 2-dimensional grid.  Each sqaure in the grid is either
    //    empty or contains a SimObject object.
    //
    //    A typical Grid is: [ [ s_0_0, s_0_1, ..., s_0_columns - 1 ],
    //                         [ s_1_0, s_1_1, ..., s_1_columns - 1 ],
    //                         ... ,
    //                         [ s_row-1_0, s_row-1_1, ..., s_row-1_columns - 1 ] ]
    //
    //    Where each s_i_j is either a SimObject object or empty.
    //

    /*@spec_public@*/ private int rows; //@invariant rows > 0;
    /*@spec_public@*/ private int columns; //@invariant columns > 0;

    // Rep Exposure warning: the objects in simobjects are exposed.
    /*@monitored@*/ private Vector simobjects; // a vector of objects of type SimObject

    /*@non_null@*/ private Random rand = new Random ();

    // Rep Invariant:
    //@invariant simobjects != null
    //@invariant simobjects.elementType == \type(SimObject)
    //@invariant simobjects.containsNull == false

    // invariant for each s: SimObject in simobjects
    //     s.isInitialized
    //     s.getRow () >= 0 && s.getRow () < rows
    //     s.getCol () >= 0 && s.getCol () < columns
    //     no other t: SimObject in simobjects the same row and column

    // Abstraction Function:
    //  AF(c) =    [ [ s_0_0, s_0_1, ..., s_0_columns - 1 ],
    //               [ s_1_0, s_1_1, ..., s_1_columns - 1 ],
    //               ... ,
    //               [ s_row-1_0, s_row-1_1, ..., s_row-1_columns - 1 ] ]
    //   where s_i_j = empty if there is no s: SimObject in simobject with s.getRow () == i and s.getCol () == j
    //         s_i_j = simobjects[k] if simobjects[k].getRow () == i and simobjects[k].getCol () == j

    Grid (int p_columns, int p_rows)
       //@requires p_rows > 0;
       //@requires p_columns > 0;
       // REQUIRES: p_rows and p_columns must be > 0
       // EFFECTS: Constructs a new grid with specified number of rows and
       //          columns.	
       // NOTE: Beware of order of rows and columns!  (x, y)
    {
        rows = p_rows;
        columns = p_columns;
        simobjects = new Vector();
	//@set simobjects.elementType = \type(SimObject)
	//@set simobjects.containsNull = false
    }
    
    public int numRows()
    // EFFECTS: Returns the number of rows in the grid.
    //@ensures \result == rows;
    {
        return rows;
    }
    
    public int numColumns()
       // EFFECTS: Returns the number of columns in the grid.
       //@ensures \result == columns;
    {
        return columns;
    }

    synchronized public void setObjectAt (int x, int y, /*@non_null@*/ SimObject newObject)
	//@requires !newObject.isInitialized
	// REQUIRES: newObject is not initialized
	// EFFECTS: places newObject on the grid at location row, col (replacing any
	//    object currently on the grid at that location).
    {
	// If an object already exists in the grid with the specified
	// row and column, remove it before adding the new one.
	simobjects.remove (getObjectAt (x, y)); //@nowarn Pre 
	newObject.init (x, y, this);
	simobjects.add (newObject); //@nowarn Exception
    } //@nowarn Exception

    public Coordinate randomFreeLocation ()
	// REQUIRES: there is at least one free square on the grid
	// EFFECTS: returns a new location on the grid that is unoccupied
	//    Note: client should lock the grid before calling this, and
	//       hold the lock until the square is used, otherwise the square
	//       may be occupied before the object is placed in it.
    {
	// Dangerous implementation...if there is no free square, this will 
	// loop forever... (but requires was not satisfied)

	while (true) {
	    int xloc = rand.nextInt (columns);
	    int yloc = rand.nextInt (rows);
	    
	    if (getObjectAt (xloc, yloc) == null) {
		return new Coordinate (xloc, yloc);
	    }
	} 
    } //@nowarn Exception

    synchronized public void setObjectAt (/*@non_null@*/ Coordinate coord, /*@non_null@*/ SimObject newObject)
	//@requires !newObject.isInitialized
	// REQUIRES: newObject is not initialized
	// EFFECTS: places newObject on the grid at location row, col (replacing any
	//    object currently on the grid at that location).
    {
	setObjectAt (coord.getX (), coord.getY (), newObject);
    }

    synchronized public void removeAllObjects ()
	// EFFECTS: removes all objects from the grid.
    {
	simobjects.setSize (0); //@nowarn Pre
    }

    synchronized public void removeObjectAt (int x, int y)
	// EFFECTS: If there is an object at (x, y), removes it.
    {
	SimObject obj = getObjectAt (x, y);

	if (obj != null) {
	    simobjects.remove (obj); 
	}
    } //@nowarn Exception

    synchronized public void removeObjectAt (/*@non_null@*/ Coordinate c)
	// EFFECTS: If there is an object at c, removes it.
    {
	removeObjectAt (c.getX (), c.getY ());
    }

    synchronized public boolean isSquareEmpty (int x, int y) throws RuntimeException
        // REQUIRES: 0 >= x < columns && 0 >= y < rows
        // EFFECTS: Returns the Cell at the specified x, y location.
	//          Returns null if nothing is present there.
    {
	return (getObjectAt (x, y) == null);
    }

    public boolean validLocation (/*@non_null@*/ Coordinate c) {
	return validLocation (c.getX (), c.getY ());
    }

    public boolean validLocation (int x, int y)
	// EFFECTS: Determines if x, y is a valid grid location.
    {
	if ((x < 0) || (x > numColumns () - 1)) {
	    return false;
	}

	if ((y < 0) || (y > numRows () - 1)) {
	    return false;
	}

	return true;
    }

    // We need the unsynchronized version of getObjectAt for the display --- we want to 
    // be able to display the grid without acquiring a lock.  This runs the risk that
    // the grid will be temporarily displayed in an inconsistent state.

    /* unsynchronized*/ public SimObject grabObjectAt (int x, int y) throws RuntimeException
        // REQUIRES: 0 >= row < rows && 0 >= col < columns
        // EFFECTS: Returns the Cell at the specified row and column.
	//          Returns null if nothing is present there.
    {
        if ((x < 0) || (x > numColumns () - 1)) {
            throw new RuntimeException ("Bad x parameter to getObjectAt: " + x);
        }

        if ((y < 0) || (y > numRows () - 1)) {
            throw new RuntimeException ("Bad y parameter to getObjectAt: " + y);
        }

	Enumeration els = simobjects.elements (); //@nowarn Race // Allow the race condition, but be careful using grab.

	while (els.hasMoreElements ()) {
	    SimObject current = (SimObject) els.nextElement ();
	    //@assume current.isInitialized // couldn't specify this invariant in ESC/Java

	    if ((current.getX () == x) && (current.getY () == y))
		{
		    return current;
		}
	}

	return null;
    }
    
    synchronized public SimObject getObjectAt (int x, int y) throws RuntimeException
        // REQUIRES: 0 >= x < columns && 0 >= y < rows
        // EFFECTS: Returns the Cell at the specified location
	//          Returns null if nothing is present there.
    {
	return grabObjectAt (x, y);
    }

    synchronized public SimObject getObjectAt (/*@non_null@*/ Coordinate c) throws RuntimeException
        // REQUIRES: 0 >= x < columns && 0 >= y < rows
        // EFFECTS: Returns the Cell at the specified location
	//          Returns null if nothing is present there.
    {
	return grabObjectAt (c.getX (), c.getY ());
    }
}
