//
// SimObject.java
//

import java.util.Enumeration;
import java.util.Vector;
import java.awt.Color;

public class SimObject {
    // OVERVIEW: A SimObject is an object than represents a simulator object.
    //    It has a grid (a circular reference to the grid containing this object),
    //    location (integer x and y), initialization state (boolean
    //    that is true if it has beein initialized).
    //
    //    A typical SimObject is < grid, x, y, initialized >.
    //
    // Note: This differs from the PS5 SimObject class: a SimObject is no longer
    // a subtype of Runnable, and it does not implement the run method.  There is
    // no thread associated with a SimObject.
    //

    // Rep:
    
    protected Grid grid;

    // subclasses have access to mx and my
    protected int mx;
    protected int my;

    //@ghost public boolean isInitialized;

    // AF(c) = < grid, mx, my, isInitialized >
    // RI(c) = 
    //         if isIntialized then 
    //              grid != null
    //              && grid.validLocation (mx, my)
    //              && grid.getObjectAt (mx, my) == this

    //@invariant !isInitialized || (grid != null)
    //  If isInitialized is true, the grid is not null.

    public SimObject ()
	// EFFECTS: Creates a new, uninitalized SimObject.
	//@ensures !isInitialized
    {
    }

    final public void init (int x, int y, /*@non_null@*/ Grid grid)
       // REQUIRES: init has not previously been called for this.  The cell is at (x, y) on grid.
       // MODIFIES: this
       // EFFECTS: Initializes the cell at x, y on grid.
       //   NOTE: This method is final, that means subtypes cannot override this method.
       //@requires !isInitialized;
       //@ensures isInitialized;
    {
        this.mx = x;
        this.my = y;
        this.grid = grid;
        //@set isInitialized = true;
    }

    public /*@non_null@*/ Color getColor () 
       // EFFECTS: Returns the yor that represents this SimObject, for
       //  painting the grid.
    {
	return Color.green; 
    } //@nowarn NonNullResult // ESC/Java spec doesn't know that Color.green is not null.

    final public int getX ()
       // REQUIRES: init has previously been called for this.
       // EFFECTS: Returns the x number that this cell is located in.
       //@requires isInitialized;
    {
        return mx;
    }

    final public Coordinate getLocation () 
       // REQUIRES: this is initialized.
       // EFFECTS: Returns the y location that this cell is located in.
       //@requires isInitialized;	
    {
	return new Coordinate (mx, my);
    }

    final public int getY ()
       // REQUIRES: this is initialized.
       // EFFECTS: Returns the y location that this cell is located in.
       //@requires isInitialized;	
    {
        return my;
    }

    //@ensures \result.elementType == \type(SimObject)
    //@requires isInitialized
    synchronized public /*@non_null@*/ Enumeration getNeighbors () {
	// REQUIRES: The grid must be locked as long as the result of getNeighbors is used.
	//    Otherwise, the neighbors could change if another object moves!
	// EFFECTS: Returns an Enumeration that yields all the objects adjacent to this
	//    in the grid.

	Vector objects = new Vector ();
	//@set objects.elementType = \type(SimObject)
	//@set objects.containsNull = false

	for (int x = mx - 1; x <= mx + 1; x++) {
	    for (int y = my - 1; y <= my + 1; y++) {
		if (x != mx || y != my) { // don't count yourself
		    if (grid.validLocation (x, y)) {
			SimObject obj = grid.getObjectAt (x, y);

			if (obj != null) {
			    objects.addElement (obj);
			}
		    }
		}
	    }
	}

	return objects.elements ();
    } //@nowarn Exception

    final public /*@non_null@*/ Grid getGrid ()
       // REQUIRES: init has previously been called for this.
       // EFFECTS: Returns the grid containing this cell.
       //    NOTE: The rep is exposed.  The Grid associated with a SimObject is part of the data abstraction.
       //@requires isInitialized == true;
    {
        return grid;
    }

}

